# MPEM toolkit v 1.0 by Thomas Barthel

import os
import collections
import shelve
import h5py


#=================================================================================
class Shelf:
  def __init__( self, fname, cacheCapacity = 128, shelfMode = 'h5py', allow_cacheGrowth = False ):
    'cacheCapacity = RAM size in MB'
    'shelfMode   = "h5py" (standard; uses cPickle for objects!=numpy.ndarray), "shelve"'
    'allow_cacheGrowth = if true, the cache grows if an incomming object is bigger than the current capacity.'
    self.fname = fname
    self.capacity = long(float(cacheCapacity)*2**20)
    self.memSize     = long(0)
    self.memSizeMax  = long(0)
    self._objMemSize = {}
    self._grow_cache = allow_cacheGrowth
    self._cache = collections.OrderedDict()
    self.mode = { 'h5py':0, 'shelve':1 }.get( shelfMode )
    if   self.mode==0:  self._shelf = h5py.File( fname, 'w' )
    elif self.mode==1:  self._shelf = shelve.open( fname, flag='n', protocol=-1 )
    
  
  def __del__(self):
    self._shelf.close()
    os.unlink( self.fname )
  
  def all_to_disk( self ):
    'Can be used to free up memory. (With time, the dictionary memory size grows up to about twice the specified memCapacity because of memory fragmentation.)'
    self.make_room( self.capacity )
    assert self.memSize == 0
    del self._cache
    self._cache = collections.OrderedDict()
  
  def memSizeMB		(self):		return float(self.memSize) / 2**20
  def memSizeMaxMB	(self):		return float(self.memSizeMax) / 2**20
  def __len__		(self):		return len(self._objMemSize)
  def __contains__	( self, key ):	return key in self._objMemSize
  def has_key		( self, key ):  return self._objMemSize.has_key( key )
  def keys	(self):			return self._objMemSize.keys()
  
  def __delitem__( self, key ):
    msize = self._objMemSize.pop( key )
    if key in self._cache:
      del self._cache[key]
      self.memSize -= msize
    if key in self._shelf:
      del self._shelf[key]
  
  def make_room( self, msize ):
    if msize > self.capacity and self._grow_cache:
      self.capacity = long(1.2*msize+1)
    assert msize <= self.capacity
    while self.memSize + msize > self.capacity:
      key, value = self._cache.popitem( last = False )
      if key not in self._shelf:
        if   self.mode==0:  self._shelf.create_dataset( key, data = value )
        elif self.mode==1:  self._shelf[key] = value
      self.memSize -= self._objMemSize[key]
  
  def __getitem__( self, key ):
    msize = self._objMemSize[ key ]
    if key not in self._cache:
      self.make_room( msize )
      self._cache[key] = self._shelf[key][()]
      self.memSize    += msize
      self.memSizeMax  = max( self.memSizeMax, self.memSize )
    return self._cache[key]
  
  def set( self, key, value, msize = -1 ):
    if msize == -1:  msize = value.nbytes
    if key in self:
      del self[key]
    self.make_room( msize )
    self._cache[key] = value
    self._objMemSize[key] = msize
    self.memSize    += msize
    self.memSizeMax  = max( self.memSizeMax, self.memSize )
  
  def __setitem__( self, key, value ):
    'available for objects that have member "nbytes"'
    self.set( key, value, value.nbytes )
