#!/usr/bin/python
# MPEM toolkit v 1.0 by Thomas Barthel

import sys
import time
import math
import psutil
import numpy as np
import networkx as nx

from generica import *
import cachedShelf
import mpem



#--------------------------------------------------------------------[ measure ]-
def measure( G, varSet, pSites, pEdges, magnSites ):
  'Evaluation of time-local edge and node observables'
  'G         = the graph'
  'varSet    = set of predefined observable labels that are to be evaluated'
  'pSites    = set of sites i for which to compute p[i](si)'
  'pEdges    = set of edges (i,j) for which to compute p[i,j](si,sj)'
  'magnSites = set of sites i for which to compute the onsite magnetization m[i]'
  ret = {}
  
  # evaluate edge and node marginals, observables and remove these marginals again
  if 'magnetization' in varSet or len(pSites) or len(magnSites):
    if 'magnetization' in varSet or len(magnSites):  assert G.graph['d']==2
    ret['normRatio_min'], ret['normRatio_max'] = mpem.eval_node_marginals( G )
  else:
    ret['normRatio_min'], ret['normRatio_max'] = mpem.eval_edge_marginals( G )
  
  M = 0
  for i in G.nodes():
    if 'p' in G.node[i]:
      di = len(G.node[i]['p'])
      if i in pSites:
        for s in range(di):
          ret['p['+str(i)+']('+str(s)+')'] = G.node[i]['p'][s]
      if di == 2:
        m = G.node[i]['p'][1] - G.node[i]['p'][0]
        M += m * mpem.NoEquivNodes( G, i )
        if i in magnSites:
          ret['m['+str(i)+']'] = m
      del G.node[i]['p']
  for i,j in G.edges():
    if 'p' in G[i][j]:
      if (i,j) in pEdges:
        di, dj = G[i][j]['p'].shape()
        for si in range(di):
          for sj in range(dj):
            ret['p['+str(i)+','+str(j)+']('+str(si)+','+str(sj)+')'] = G[i][j]['p'][si,sj]
      del G[i][j]['p']
  M /= mpem.totalNoNodes( G )
  if 'magnetization' in varSet:
    ret['magnetization'] = M
  
  return ret


#--------------------------------------------------------------[ measureCorrel ]-
def measureCorrel( G, varSet, pSites, pEdges ):
  'Evaluation of temporal edge and node correlations'
  'G         = the graph'
  'varSet    = set of predefined observable labels that are to be evaluated'
  'pSites    = set of sites i for which to compute p[i](si(t),si(s))'
  'pEdges    = set of edges (i,j) for which to compute p[i,j](si(t),sj(t),si(s),sj(s))'
  for i,j in G.edges():
    if 'A' in G[i][j]:  T=len(G[i][j]['A'])-2; break
  ret = [{} for n in range(T)]
  assert T>0
  
  # evaluate edge and node marginals, observables and remove these marginals again
  if len(pSites):
    mpem.eval_node_correlMarginals( G )
  else:
    mpem.eval_edge_correlMarginals( G )
  
  for i in G.nodes():
    if 'pCorrels' in G.node[i]:
      Ns, dI, di = G.node[i]['pCorrels'].shape
      assert Ns==T and dI==di
      if i in pSites:
        for t in range(T):
         for Si in range(di):
          for si in range(di):
            ret[t]['p['+str(i)+']('+str(Si)+','+str(si)+')'] = G.node[i]['pCorrels'][t,Si,si]
      del G.node[i]['pCorrels']
  for i,j in G.edges():
    if 'pCorrel' in G[i][j]:
      if (i,j) in pEdges:
        Ns, dI, dJ, di, dj = G[i][j]['pCorrels'].shape
        assert Ns==T and dI==di and dJ==dj
        for t in range(T):
         for Si in range(di):
          for Sj in range(dj):
           for si in range(di):
            for sj in range(dj):
              ret[t]['p['+str(i)+','+str(j)+']('+str(Si)+','+str(Sj)+','+str(si)+','+str(sj)+')'] = G[i][j]['p'][t,Si,Sj,si,sj]
      del G[i][j]['pCorrels']
  
  return ret


#---------------------------------------------------------------[ show_measure ]-
def show_measure( outfile, t, outVar, val, tic, shelf ):
  val['t']           = t
  val['runtime_cpu'] = time.clock()-tic
  val['mem_total']   = process.get_memory_info()[0] / float(2**20)
  val['shelf_memSizeMax'] = 0
  if shelf != None:  val['shelf_memSizeMax'] = shelf.memSizeMaxMB()
  line = string.join( ['%.12g'%val[var] for var in outVar], '\t')
  print >> outfile, line
  outfile.flush()
  print string.join( outVar, '\t' )
  print line

#---------------------------------------------------------[ show_measureCorrel ]-
def show_measureCorrel( outfile, outVar, val ):
  Val = {}
  Val['t'] = len(val)
  print string.join( outVar, '\t' )
  for s in range(len(val)):
    Val['s'] = s
    Val.update( val[s] )
    line = string.join( ['%.12g'%Val[var] for var in outVar], '\t')
    print >> outfile, line
    print line
  outfile.flush()



#=================================================================================
#========================================================================={ MAIN ]
program = sys.argv[0].rsplit('.py',1)[0].rsplit('/',1)[-1]
process = psutil.Process(os.getpid())

in_model = ''
tMax = 7
algo = 'densityMatrixTruncation'
truncator = mpem.Truncator()
outVar     = []
pSites     = set([])
pEdges     = set([])
magnSites  = set([])
outVarCorrel = []
correlSites  = set([])
correlEdges  = set([])
shelfFile      = '/usr/tmp/'+program+'.hdf5.[unique_id]'
shelfCacheSize = 512

print '\n >>>> Matrix product edge message simulation <<<<  v 1.0, T. Barthel'
print 'usage: ' + program + '.py'
print '\t*  -in_model <*.model.bin>  (-tMax <.='+str(tMax)+'>)'
print '\t  (-algorithm <.="'+algo+'">)'
print '\t* (-truncator_statesMin <.='+str(truncator['statesMin'])+'>)  (-truncator_statesMax <.='+str(truncator['statesMax'])+'>)'
print '\t  (-truncator_dont_normalize)  (-truncator_errorMax <.='+str(truncator['errorMax'])+'>)  (-truncator_schmidtTruncMax <.='+str(truncator['schmidtTruncMax'])+'>)'
print '\t* (-prob_onsite <vertex> ..)  (-prob_edge <vertex i> <vertex j> ..)'
print '\t  (-measure_magn_onsite <vertex> ..)  (-out_col <variable label> ..)'
print '\t* (-correl_onsite <vertex> ..)  (-correl_edge <vertex i> <vertex j> ..)  (-outCorrel_col <variable label> ..)'
print '\t* (-shelfFile <file name ='+shelfFile+'>)  (-shelfCache <.='+str(shelfCacheSize)+'MB>)'
print ''
print ' Possible algorithms are:  naiveConstruction, orthoConstruction, densityMatrixTruncation.'
print ' For the moment, a shelfFile is only employed, for algorithm densityMatrixTruncation and for the evaluation of temporal correlators.'
print ' Available -out_col variables:  t, (magnetization), (m[i]), (p[i](s)), (p[i,j](si,sj)), (normRatio_min), (normRatio_max), statesKept_max, schmidtTrunc_max, error_max, error_sum_max, cost_max, cost_sum, runtime_cpu, mem_total, shelf_memSizeMax'
print ' Available -outCorrel_col variables:  t, s, (p[i](si(t),si(s))), (p[i,j](si(t),sj(t),si(s),sj(s)))'
print ''


# ----------------------------------------- read command line parameters
modable = set(['-prob_onsite','-prob_edge','-measure_magn_onsite','-out_col',\
               '-correl_onsite','-correl_edge','-outCorrel_col'])
mode = 'none'
n = 1

while n < len(sys.argv) and ( is_cmdlineArg( sys.argv[n] ) or mode!='none' ):
  arg = sys.argv[n]
  [dn, mode, switch] = read_cmdlineArg( arg, mode, modable )
  n += dn
  
  if switch == '-in_model':
    in_model = sys.argv[n]
  elif switch == '-tMax':
    tMax = int(sys.argv[n])
  elif switch == '-algorithm':
    algo = sys.argv[n]
  elif switch == '-truncator_statesMin':
    truncator['statesMin'] = int(sys.argv[n])
  elif switch == '-truncator_statesMax':
    truncator['statesMax'] = int(sys.argv[n])
  elif switch == '-truncator_errorMax':
    truncator['errorMax'] = float(sys.argv[n])
  elif switch == '-truncator_schmidtTruncMax':
    truncator['schmidtTruncMax'] = float(sys.argv[n])
  elif switch == '-truncator_schmidtTruncMax':
    truncator['schmidtTruncMax'] = float(sys.argv[n])
  elif switch == '-prob_onsite':
    pSites.add( int(sys.argv[n]) )
  elif switch == '-prob_edge':
    pEdges.add( ( int(sys.argv[n]),int(sys.argv[n+1]) ) )
    n += 1
  elif switch == '-measure_magn_onsite':
    magnSites.add( int(sys.argv[n]) )
  elif switch == '-out_col':
    outVar.append( sys.argv[n] )
  elif switch == '-correl_onsite':
    correlSites.add( int(sys.argv[n]) )
  elif switch == '-correl_edge':
    correlEdges.add( ( int(sys.argv[n]),int(sys.argv[n+1]) ) )
    n += 1
  elif switch == '-outCorrel_col':
    outVarCorrel.append( sys.argv[n] )
  elif switch == '-shelfFile':
    shelfFile = sys.argv[n]
  elif switch == '-shelfCache':
    shelfCacheSize = float(sys.argv[n])
  else:
    print '\nUnknown switch ',switch
    sys.exit(2)
  n += 1

if len(in_model) == 0:
  print ''
  sys.exit(2)


# ----------------------------------------- READ MODEL
G = nx.read_gpickle( in_model ).to_undirected()
G = nx.DiGraph(G)
d = G.graph['d']

print 'd =',d
print 'G.number_of_nodes() =',G.number_of_nodes()
print 'totalNoNodes =',mpem.totalNoNodes( G )
print 'zMax =',max([G.out_degree(i) for i in G.nodes()])


# ----------------------------------------- INIT SHELF
G.graph['shelf'] = None
if algo == 'densityMatrixTruncation':
  shelfFile = create_unique_file( shelfFile )
  G.graph['shelf'] = cachedShelf.Shelf( shelfFile, shelfCacheSize, shelfMode='h5py', allow_cacheGrowth = True )


# ----------------------------------------- OUTPUT FILES
outVarSet       = set(outVar)
outVarCorrelSet = set(outVarCorrel)

fname = in_model.rsplit('.model.bin')[0] + '~MPEM'
truncLabel = mpem.truncatorLabel( truncator )
if len(truncLabel):  fname += '_'+truncLabel

if len(outVar):
  fname_stat = fname + '.expect.table.txt'
  outfile_stat = open( fname_stat, 'w' )
  print 'outfile_stat =', fname_stat
  print >> outfile_stat, '# ' + string.join( outVar, '\t' )

if len(outVarCorrel):
  fname_correl = fname + '.correl.table.txt'
  outfile_correl = open( fname_correl, 'w' )
  print 'outfile_correl =', fname_correl
  print >> outfile_correl, '# ' + string.join( outVarCorrel, '\t' )


# ----------------------------------------- initial state (t=0)
mpem.initialize( G )
tic = time.clock()

print '\n======== t = 0 ================================================'
truncInfo = mpem.truncationInfoGlobal_init()
truncInfo.update( measure( G, outVarSet, pSites, pEdges, magnSites ) )
show_measure( outfile_stat, 0, outVar, truncInfo, tic, G.graph['shelf'] )


# ----------------------------------------- loop for t>0
for t in range(0,tMax-1):
  print '\n======== t = '+str(t)+' -> '+str(t+1)+' ================================================'
  truncInfo = {}
  mpem.timeStep( G, truncator, truncInfo, algo )
  
  truncInfo.update( measure( G, outVarSet, pSites, pEdges, magnSites ) )
  if len(outVar):
    show_measure( outfile_stat, t+1, outVar, truncInfo, tic, G.graph['shelf'] )
  if len(outVarCorrel):
    correl = measureCorrel( G, outVarCorrelSet, correlSites, correlEdges )
    show_measureCorrel( outfile_correl, outVarCorrel, correl )


if len(outVar):
  outfile_stat.close()
if len(outVarCorrel):
  outfile_correl.close()
print 'elapsed time =', time.clock()-tic
