# MPEM toolkit v 1.0 by Thomas Barthel

from math import sqrt
from copy import deepcopy
import numpy as np
import numpy.linalg as linalg
import networkx as nx

from transitionMatrix import *
from truncation import *


#__________________________________________________________________________________________
#========================================================================= Initialization =

#------------------------------------------------------------[ initialize_edge ]-
def initialize_edge( G, i, j ):
  di = dimension_i(G,i)
  G[i][j]['A'] = [ np.ones((1,1,1,1)), np.zeros((di,1,1,1)) ]	# A[t][si(t-1),sj(t),a,b]
  for si in range(di):
    G[i][j]['A'][1][si,0,0,0] = G.node[i]['P'][si]

#-----------------------------------------------------------------[ initialize ]-
def initialize( G ):
  if 'equivTo' not in G.node[0]:	# graph without equivalency information
    for i,j in G.edges():
      initialize_edge( G, i, j )
  else:					# graph with equivalent edges
    for i,j in G.edges():
      if G.node[i]['equivTo']==i and G.node[j]['equivTo'] in [i,j]:
        initialize_edge( G, i, j )


#__________________________________________________________________________________________
#============================================================================== marginals =

#--------------------------------------------------------------[ edge_marginal ]-
def edge_marginal( A, Ainv ):
  'Given edge messages mu_{i->j}(t) and mu_{j->i}(t), this function computes the edge marginal p(si(t),sj(t)) at time t.'
  T = len(A)-2
  assert len(Ainv) == T+2
  assert A[0]   .shape[0] == 1 and A[0]   .shape[2] == 1
  assert Ainv[0].shape[0] == 1 and Ainv[0].shape[2] == 1
  E = np.einsum( 'jb,id->ijbd', A[0][0,:,0,:], Ainv[0][0,:,0,:] )
  factor = 1
  for t in range(1,T+1):
    factor*=linalg.norm(E)
    E *= 1./linalg.norm(E)
    E = np.einsum( 'ijac,jIcd->iIad', E, Ainv[t] )
    E = np.einsum( 'iJab,iIad->IJbd', A[t], E )
  assert E.shape[:2]==(1,1)
  assert A[T+1]   .shape[1] == 1 and A[T+1]   .shape[3] == 1
  assert Ainv[T+1].shape[1] == 1 and Ainv[T+1].shape[3] == 1
  E = np.einsum( 'ac,jc->ja', E[0,0,:,:], Ainv[T+1][:,0,:,0] )
  p = np.einsum( 'ia,ja->ij', A[T+1][:,0,:,0], E )
  return 1./p[:].sum() * p, factor/p[:].sum()


#-------------------------------------------------------[ edge_correlMarginals ]-
def edge_correlMarginals( A, Ainv, shelf ):
  'Given edge messages mu_{i->j}(t) and mu_{j->i}(t), this function computes the edge correlator marginals p(si(t),sj(t),si(s),sj(s)) for time t and all times s<t.'

  T = len(A)-2
  assert len(Ainv) == T+2
  assert A[0]   .shape[0] == 1 and A[0]   .shape[2] == 1
  assert Ainv[0].shape[0] == 1 and Ainv[0].shape[2] == 1
  EL = lambda s: 'E'+str(s)		# norm matrices for the left blocks
  factor = 1
  normRatio = T*[-1,]
  E = np.einsum( 'jb,id->ijbd', A[0][0,:,0,:], Ainv[0][0,:,0,:] )
  for t in range(1,T+1):
    shelf[EL(t-1)], normRatio[t-1] = E, factor
    factor*=linalg.norm(E)
    E *= 1./linalg.norm(E)
    E = np.einsum( 'ijac,jIcd->iIad', E, Ainv[t] )
    E = np.einsum( 'iJab,iIad->IJbd', A[t], E )
  assert E.shape[:2]==(1,1)
  
  pCorrels = T*[0,]
  normRatio_min, normRatio_max = [1e8,0]
  assert A[T+1]   .shape[1] == 1 and A[T+1]   .shape[3] == 1
  assert Ainv[T+1].shape[1] == 1 and Ainv[T+1].shape[3] == 1
  factor = 1
  F = np.einsum( 'Kjb,Lid->KLijbd', A[T+1][:,:,:,0], Ainv[T+1][:,:,:,0] )
  for t in range(1,T+1)[::-1]:
    factor*=linalg.norm(F)
    F *= 1./linalg.norm(F)
    F = np.einsum( 'KLijbd,Jicd->KLJjbc', F, Ainv[t] )
    F = np.einsum( 'Ijab,KLJjbc->KLIJac', A[t], F )
    pCorrels[t-1] = np.einsum( 'ijac,KLijac->KLij', shelf[EL(t-1)], F )
    pCorrels[t-1], nr = 1./pCorrels[t-1][:].sum()*pCorrels[t-1], factor*normRatio[t-1]/pCorrels[t-1][:].sum()
    normRatio_min = min(normRatio_min,nr)
    normRatio_max = max(normRatio_max,nr)
    del shelf[EL(t-1)]	# free up memory
  
  return pCorrels, normRatio_min, normRatio_max


#--------------------------------------------------------[ eval_edge_marginals ]-
def eval_edge_marginals( G ):
  noEquivInfo = 'equivTo' not in G.node[0]
  normRatio_min, normRatio_max = [1e8,0]
  for i,j in G.edges():
    if 'p' in G[i][j]:  del G[i][j]['p']
  for i,j in G.edges():
    if 'p' not in G[i][j] and (noEquivInfo or (G.node[i]['equivTo']==i and G.node[j]['equivTo'] in [j,i])):
      if noEquivInfo or G.node[j]['equivTo']==j:
        Ainv = G[j][i]['A']
      else:
        Ainv = G[i][j]['A']
      p_ij, nr = edge_marginal( G[i][j]['A'], Ainv )
      normRatio_min = min(normRatio_min,nr)
      normRatio_max = max(normRatio_max,nr)
      G[i][j]['p'] = p_ij
      G[j][i]['p'] = p_ij.transpose( 1, 0 )
  return normRatio_min, normRatio_max


#--------------------------------------------------[ eval_edge_correlMarginals ]-
def eval_edge_correlMarginals( G ):
  noEquivInfo = 'equivTo' not in G.node[0]
  normRatio_min, normRatio_max = [1e8,0]
  for i,j in G.edges():
    if 'pCorrels' in G[i][j]:  del G[i][j]['pCorrels']
  for i,j in G.edges():
    if 'pCorrels' not in G[i][j] and (noEquivInfo or (G.node[i]['equivTo']==i and G.node[j]['equivTo'] in [j,i])):
      if noEquivInfo or G.node[j]['equivTo']==j:
        Ainv = G[j][i]['A']
      else:
        Ainv = G[i][j]['A']
      p_ij, nrMin, nrMax = edge_correlMarginals( G[i][j]['A'], Ainv, G.graph['shelf'] )
      normRatio_min = min(normRatio_min,nrMin)
      normRatio_max = max(normRatio_max,nrMax)
      p_ij = np.array( p_ij )
      G[i][j]['pCorrels'] = p_ij
      G[j][i]['pCorrels'] = p_ij.transpose( (0,2,1,4,3) )
  return normRatio_min, normRatio_max


#--------------------------------------------------------[ eval_node_marginals ]-
def eval_node_marginals( G ):
  noEquivInfo = 'equivTo' not in G.node[0]
  normRatio_min, normRatio_max = eval_edge_marginals( G )
  for i in G.nodes():
    if 'p' in G[i]:  del G[i]['p']
  for i in G.nodes():
    if (noEquivInfo or G.node[i]['equivTo']==i):
      p = np.zeros( dimension_i( G, i ) )
      for j in G.neighbors(i):
        j0 = j
        if not noEquivInfo and G.node[j]['equivTo'] != i:
          j0 = G.node[j]['equivTo']
        p += np.einsum( 'ij->i', G[i][j0]['p'] )
      p *= 1./len(G.neighbors(i))
      G.node[i]['p'] = p
  return normRatio_min, normRatio_max


#--------------------------------------------------[ eval_node_correlMarginals ]-
def eval_node_correlMarginals( G ):
  noEquivInfo = 'equivTo' not in G.node[0]
  normRatio_min, normRatio_max = eval_edge_correlMarginals( G )
  for i,j in G.edges():
    if 'pCorrels' in G[i][j]:  T=len(G[i][j]['pCorrels']); break
  for i in G.nodes():
    if 'pCorrels' in G[i]:  del G[i]['pCorrels']
  for i in G.nodes():
    if (noEquivInfo or G.node[i]['equivTo']==i):
      p = np.zeros( (T,dimension_i(G,i),dimension_i(G,i)) )
      for j in G.neighbors(i):
        j0 = j
        if not noEquivInfo and G.node[j]['equivTo'] != i:
          j0 = G.node[j]['equivTo']
        p += np.einsum( 'sIJij->sIi', G[i][j0]['pCorrels'] )
      p *= 1./len(G.neighbors(i))
      G.node[i]['pCorrels'] = p
  return normRatio_min, normRatio_max


#__________________________________________________________________________________________
#================================================================================ C_from* =

#------------------------------------------------------------------[ C_from_Aw ]-
def C_from_Aw( A, w ):
  'Naive construction of C tensor for t>0 (bulk).'
  d = w.shape
  Z = len(A)
  if Z==0:  return np.array(w).reshape(d[0],d[1],1,1)
  C = d[0]*[0]
  for si in range(d[0]):
    n = 1
    wA = w[si,:].reshape( d[1], d[n+1], np.prod(d[(n+2):]), 1, 1 )
    for k in range(Z):
      Si = min( si, A[k].shape[1]-1 )	# takes care of fake indices
      wA = np.einsum( 'jeEAB,eab->jEAaBb', wA, A[k][:,Si,:] )
      if n < Z:
        n += 1
        dims = wA.shape
        wA = wA.reshape( d[1], d[n+1], dims[1]/d[n+1], dims[2]*dims[3], dims[4]*dims[5] )
    dims = wA.shape
    wA = wA.reshape( d[1], dims[2]*dims[3], dims[4]*dims[5] )
    C[si] = wA
  return np.array(C)


#-----------------------------------------------------------------[ C_from_AwX ]-
def C_from_AwX( A, w, X ):
  'Construction of C tensor including tensor X on the right, originating from the truncation of the C tensor for t+1.'
  'REMARK: Doing the contractions from big to small t (L<-R), as done here, is also useful for the case where w(si(t)|sj(t-1)...) also depends on si(t-1). This si(t-1) label can then be transported in X to the next contraction for t-1.'
  d = w.shape
  Dc = X.shape[1]
  Z = len(A)
  Da = [A[k].shape[2] for k in range(Z)]
  Db = [A[k].shape[3] for k in range(Z)]
  assert X.ndim == 2 and X.shape[0] == np.prod(Db)
  if Z==0:  return np.einsum( 'ijab,bc->ijac', np.array(w).reshape(d[0],d[1],1,1), X )
  C = d[0]*[0]
  for si in range(d[0]):
    n = 1
    AwX = np.outer( w[si,:], X )
    AwX = AwX.reshape( d[1], d[n+1], np.prod(d[(n+2):]), 1, Db[n-1], np.prod(Db[n:]), Dc )
    for k in range(Z):
      Si = min( si, A[k].shape[1]-1 )	# takes care of fake indices
      AwX = np.einsum( 'eab,jeEAbBc->jEAaBc', A[k][:,Si,:], AwX )
      if n < Z:
        n += 1
        dims = AwX.shape
        AwX = AwX.reshape( d[1], d[n+1], dims[1]/d[n+1], dims[2]*dims[3], Db[n-1], dims[4]/Db[n-1], Dc )
    dims = AwX.shape
    AwX = AwX.reshape( d[1], dims[2]*dims[3], Dc )
    C[si] = AwX
  return np.array(C)


#-----------------------------------------------------------------[ C_from_XwA ]-
def C_from_XwA( X, w, A ):
  'Construction of C tensor including tensor X on the left, originating from the truncation of the C tensor for t-1.'
  d = w.shape
  Dc = X.shape[0]
  Z = len(A)
  Db = [A[k].shape[2] for k in range(Z)]
  Da = [A[k].shape[3] for k in range(Z)]
  assert X.ndim == 2 and X.shape[1] == np.prod(Db)
  if Z==0:  return np.einsum( 'cb,ijba->ijca', X, np.array(w).reshape(d[0],d[1],1,1) )
  C = d[0]*[0]
  for si in range(d[0]):
    n = 1
    XwA = np.outer( X, w[si,:] )
    XwA = XwA.reshape( Dc, Db[n-1], np.prod(Db[n:]), d[1], d[n+1], np.prod(d[(n+2):]), 1 )
    for k in range(Z):
      Si = min( si, A[k].shape[1]-1 )	# takes care of fake indices
      XwA = np.einsum( 'cbBjeEA,eba->cBjEAa', XwA, A[k][:,Si,:] )
      if n < Z:
        n += 1
        dims = XwA.shape
        XwA = XwA.reshape( Dc, Db[n-1], dims[1]/Db[n-1], d[1], d[n+1], dims[3]/d[n+1], dims[4]*dims[5] )
    dims = XwA.shape
    XwA = XwA.reshape( Dc, d[1], dims[4]*dims[5] )
    C[si] = np.transpose( XwA, [1,0,2] )
  return np.array(C)



#__________________________________________________________________________________________
#================================================================================ F_from* =

#-----------------------------------------------------------------[ F_from_AwF ]-
def F_from_AwF( A, w, F ):
  'Construction of (right block) norm matrix F[t] from the A matrices and w at time t and the previous norm matrix F[t+1].'
  'This corresponds to F[t] = \sum_{si,sj} C[t](si,sj)*F[t+1]*herm(C[t](si,sj)).'
  'The computation cost is M^(2Z+1)=M^(2z-1) where Z is the number of involved A matrices.'
  d = w.shape
  Z = len(A)
  Da = [A[k].shape[2] for k in range(Z)]
  Db = [A[k].shape[3] for k in range(Z)]
  Dc = np.prod(Db)
  assert F.ndim == 2 and F.shape == (Dc,Dc)
  if Z==0:
    wF = np.einsum( 'ijab,bc->ijac', np.array(w).reshape(d[0],d[1],1,1), F )
    return np.einsum( 'ijac,ijdc->ad', wF, np.array(w).reshape(d[0],d[1],1,1) )
  
  Ret = np.zeros( 2*(np.prod(Da),) )
  for si in range(d[0]):
   for sj in range(d[1]):	# earlier we let numpy do this sum ... but that required more memory
    
    # 1) contract w and A on left
    Dc = np.prod(Db)
    n = 1
    Si = min( si, A[n-1].shape[1]-1 )	# takes care of fake indices
    wA = np.einsum( 'eab,eE->Eab', A[n-1][:,Si,:], w[si,sj,:].reshape( d[n+1], np.prod(d[(n+2):]) ) )
    wAF = np.einsum( 'Eab, bBc->EaBc', wA, F.reshape( Db[n-1], np.prod(Db[n:]), Dc ) );  del wA
    if n < Z:
      n += 1;  dims = wAF.shape
      wAF = wAF.reshape( d[n+1], dims[0]/d[n+1], dims[1], Db[n-1], dims[2]/Db[n-1], Dc )
    for k in range(1,Z):
      Si = min( si, A[n-1].shape[1]-1 )	# takes care of fake indices
      wAF = np.einsum( 'eab,eEAbBc->EAaBc', A[k][:,Si,:], wAF )
      if n < Z:
        n += 1;  dims = wAF.shape
        wAF = wAF.reshape( d[n+1], dims[0]/d[n+1], dims[1]*dims[2], Db[n-1], dims[3]/Db[n-1], Dc )
    dims = wAF.shape
    wAF = wAF.reshape( dims[1]*dims[2], Dc )
    
    # 2) contract w and A on right
    Dc = np.prod(Da)
    n = 1
    wAF = wAF.transpose( 1, 0 )		# transpose to then use code analogous to above
    Si = min( si, A[n-1].shape[1]-1 )	# takes care of fake indices
    wA = np.einsum( 'eab,eE->Eab', A[n-1][:,Si,:], w[si,sj,:].reshape( d[n+1], np.prod(d[(n+2):]) ) )
    wAF = np.einsum( 'Eab, bBc->EaBc', wA, wAF.reshape( Db[n-1], np.prod(Db[n:]), Dc ) );  del wA
    if n < Z:
      n += 1;  dims = wAF.shape
      wAF = wAF.reshape( d[n+1], dims[0]/d[n+1], dims[1], Db[n-1], dims[2]/Db[n-1], Dc )
    for k in range(1,Z):
      Si = min( si, A[n-1].shape[1]-1 )	# takes care of fake indices
      wAF = np.einsum( 'eab,eEAbBc->EAaBc', A[k][:,Si,:], wAF )
      if n < Z:
        n += 1;  dims = wAF.shape
        wAF = wAF.reshape( d[n+1], dims[0]/d[n+1], dims[1]*dims[2], Db[n-1], dims[3]/Db[n-1], Dc )
    dims = wAF.shape
    wAF = wAF.reshape( dims[1]*dims[2], Dc )
    
    Ret += np.transpose( wAF )	# in principle, "transpose" is superfluous, as wAF should be hermitian
  
  return Ret


#__________________________________________________________________________________________
#======================================================================== densityMatrix_* =

#---------------------------------------------------------[ densityMatrix_left ]-
# Construction of (left block) density matrix rho[t] from X (the result of the last truncation on the left), the A matrices and w at time t and the norm matrix F[t+1] for the right block.
# This corresponds to rho[t](si,sj|ri,rj) = X*C[t](si,sj)*F[t+1]*herm(C[t](ri,rj))*herm(X).
# The computation cost is M^(2Z+1)=M^(2z-1) where Z is the number of involved A matrices.
def densityMatrix_left( X, w, A, F ):
  C = C_from_XwA( X, w, A )
  rho = np.einsum( 'ijab,bc->ijac', C, F )
  rho = np.einsum( 'ijac,IJdc->ijaIJd', rho, C )
  return rho, C


#__________________________________________________________________________________________
#======================================================================== truncationSweep =

#------------------------------------------------------[ truncationSweep_right ]-
def truncationSweep_right( C, trunc, infoSweep ):
  'sweep L->R truncating'
  T = len(C)-3
  X = np.ones((1,1))
  for t in range(0,T+1):
    Shape = C[t].shape
    C[t] = np.einsum( 'ka,aijb->kijb', X, C[t] ).reshape( -1, C[t].shape[3] )
    C[t],L,V, info = truncate( C[t], trunc )
    truncationInfoSweep_update( infoSweep, info )
    C[t] = C[t].reshape(-1,Shape[1],Shape[2],len(L))
    X = np.einsum('k,kb->kb',L,V)
  C[T+1] = np.einsum( 'ka,aijb->kijb', X, C[T+1] )


#__________________________________________________________________________________________
#================================================================================= evolve =

#----------------------------------------------------------------[ evolve_edge ]-
def evolve_edge( P, w, A, A_neighb, trunc, algo='orthoConstruction', shelf=None ):
  'Evolves the edge message mu_{i->j} for one time step.'
  'P = the t=0 state probabilities of vertex i.'
  'w = the tranfer matrix for vertex i.'
  'A        = the MPEM for i->j'
  'A_neighb = a list of MPEMs for k->i for all neighbors k!=j of i.'
  'trunc    = truncation parameters'
  'algo     = "naiveConstruction", "orthoConstruction", or "densityMatrixTruncation"'
  
  T = len(A)-2
  di = w.shape[0]
  dj = w.shape[1]
  Z = len(A_neighb)
  infoSweep = {}	# will contain summary of truncation information
  C = (T+3)*[0]
  noTrunc = NoTruncator()
  
  if algo == 'naiveConstruction':
    # (a) apply transfer matrices w -> from A to C
    for t in range(0,T+2):
      if t==0 :  C[0] = C_from_Aw( [A_neighb[k][t] for k in range(Z)], P.reshape((di,1)+Z*(1,)) )
      else:      C[t] = C_from_Aw( [A_neighb[k][t] for k in range(Z)], w )
    print 'dims_aN =', [C[t].shape[3] for t in range(T+2)]
    
    # (a2) sweep L<-R (no truncations) to orthonormalize the C matrices
    X = np.ones((1,1))
    for t in range(1,T+2)[::-1]:
      C[t] = np.einsum( 'ijab,bk->aijk', C[t], X ).reshape( C[t].shape[2], -1 )
      U,L,C[t], info = truncate( C[t], noTrunc )
      truncationInfoSweep_update( infoSweep, info )
      C[t] = C[t].reshape(len(L),di,dj,-1)
      X = np.einsum('ak,k->ak',U,L)
    C[0] = np.einsum( 'ijab,bk->aijk', C[0], X )
    print 'L<-R. dims_aN2 =', [C[t].shape[3] for t in range(0,T+2)]
  
    # (b) sweep L->R truncating
    truncationSweep_right( C, trunc, infoSweep )
    print 'L->R. dims_b =', [C[t].shape[3] for t in range(0,T+2)]

  elif algo == 'orthoConstruction':
    # (a) Going L<-R, construct C from A and w and orthogonalize directly. In this way, the C matrices have the full tensor product basis only on the left; zero singular values are already truncated on the right. This reduces costs substantially.
    # REMARK: Doing the contractions from big to small t (L<-R), as done here, is also useful for the case where w(si(t)|sj(t-1)...) also depends on si(t-1). This si(t-1) label can then be transported in X to the next contraction for t-1.
    X = np.ones((1,1))
    for t in range(1,T+2)[::-1]:
      C[t] = C_from_AwX( [A_neighb[k][t] for k in range(Z)], w, X )
      C[t] = np.einsum('ijab->aijb',C[t]).reshape( C[t].shape[2], -1 )
      U,L,C[t], info = truncate( C[t], noTrunc )
      truncationInfoSweep_update( infoSweep, info )
      C[t] = C[t].reshape(len(L),di,dj,-1)
      X = np.einsum('ak,k->ak',U,L)
    C[0] = C_from_Aw( [A_neighb[k][0] for k in range(Z)], P.reshape((di,1)+Z*(1,)) )
    C[0] = np.einsum( 'ijab,bk->aijk', C[0], X )
    print 'L<-R. dims_aOC =', [C[t].shape[3] for t in range(0,T+2)]
  
    # (b) sweep L->R truncating
    truncationSweep_right( C, trunc, infoSweep )
    print 'L->R. dims_b =', [C[t].shape[3] for t in range(0,T+2)]
  
  elif algo == 'densityMatrixTruncation':
    # (a) Going L->R, we construct C from A and w and orthogonalize+truncate directly, based on the normMatrix for the right block (all computed in a previous sweep L<-R). In this way, the C matrices have the full tensor product basis only on the right; the left is already truncated. This reduces costs substantially to M^(2Z+1)=M^(2z-1).
    F = lambda s: 'F'+str(s)		# norm matrices for the right blocks
    shelf[F(T+2)] = np.ones((1,1))
    for t in range(1,T+2)[::-1]:
      shelf[F(t)] = F_from_AwF( [A_neighb[k][t] for k in range(Z)], w, shelf[F(t+1)] )
    del shelf[F(T+2)]
    X = np.ones((1,1))
    for t in range(0,T+1):
      W = w if t>0 else P.reshape((di,1)+Z*(1,))
      rho, XwA = densityMatrix_left( X, W, [A_neighb[k][t] for k in range(Z)], shelf[F(t+1)] )
      del shelf[F(t+1)]	# free up memory
      d0, d1 = rho.shape[:2]
      C[t],R, info = truncate_dm( rho.reshape( d0*d1*X.shape[0], -1 ), trunc )
      truncationInfoSweep_update( infoSweep, info )
      C[t] = np.transpose( C[t].reshape( d0, d1, X.shape[0], -1 ), [2,0,1,3] )
      X = np.einsum( 'bija,ijbc->ac', C[t], XwA )
    C[T+1] = C_from_XwA( X, w, [A_neighb[k][T+1] for k in range(Z)] )
    C[T+1] = np.transpose( C[T+1], [2,0,1,3] )
    print 'L->R. dims_abDMT =', [C[t].shape[3] for t in range(0,T+2)]
  
  else:
    print 'Unknown construction algorithm =', algo
    assert False
  
  
  # (c) sweep L<-R, truncating and reordering indices
  #       (si_0)(si_1,sj_0)(si_2,sj_1) ... (si_{T+1},sj_T)
  #     ->(sj_0,si_0)(sj_1,si_1)(sj_2,si_2) ... (si_{T+1})
  X = np.ones((1,1,1))
  for t in range(1,T+2)[::-1]:
    C[t] = np.einsum( 'aijb,Jbk->jaJik', C[t], X ).reshape( dj*C[t].shape[0], -1 )
    U,L,C[t], info = truncate( C[t], trunc )
    truncationInfoSweep_update( infoSweep, info )
    C[t] = C[t].reshape(len(L),X.shape[0],di,-1)
    X = np.einsum('Ak,k->Ak',U,L).reshape(dj,-1,len(L))
  C[0] = np.einsum( 'aib,Jbk->aJik', C[0][:,:,0,:], X )
  print 'L<-R. dims_c =', [C[t].shape[3] for t in range(0,T+2)]
  
  # (d) sweep L->R, truncating and reordering indices to normal form
  #        (sj_0,si_0)(sj_1,si_1)(sj_2,si_2) ... (si_{T+1})
  #     -> (sj_0)(si_0,sj_1)(si_1,sj_2) ... (si_{T-1},sj_T)(si_T)(si_{T+1})
  X = np.ones((1,1,1))
  for t in range(0,T+1):
    C[t] = np.einsum( 'kIa,ajib->Ijkib', X, C[t] ).reshape( -1, di*C[t].shape[3] )
    C[t],L,V, info = truncate( C[t], trunc )
    truncationInfoSweep_update( infoSweep, info )
    C[t] = C[t].reshape(X.shape[1],dj,-1,len(L))
    X = np.einsum('k,kB->kB',L,V).reshape(len(L),di,-1,)
  C[T+2] = C[T+1].copy()
  C[T+1],L,V, info = truncate( X.reshape( X.shape[0]*di, -1 ), trunc )
  truncationInfoSweep_update( infoSweep, info )
  C[T+1] = np.einsum( 'aijk->ijak', C[T+1].reshape(-1,di,1,len(L)) )
  X = np.einsum('k,kb->kb',L,V)
  C[T+2] = np.einsum( 'ka,ajib->ijkb', X, C[T+2] )
  print 'L->R. dims_d =', [C[t].shape[3] for t in range(T+3)]
  
  return C, infoSweep


#-------------------------------------------------------------------[ timeStep ]-
def timeStep( G, trunc, truncInfoGlobal, algo='orthoConstruction' ):
  'A single time step in the recursive construction of edge messages.'
  
  # Evolve all edge messages A->B
  if 'equivTo' not in G.node[0]:	# graph without equivalency information
    for i,j in G.edges():
      print ' --- i,j=',i,j,'---------'
      k_minus_j = neighb_minus_j( G, i, j )
      wSorted   = w_sorted( G, i, j )
      A_neighb = [ G[k][i]['A'] for k in k_minus_j ]
      # Assign new A matrices to edge, but call them 'B' for the moment as the old 'A' matrices are in general still needed for updates of other edges
      G[i][j]['B'], infoSweep = evolve_edge( G.node[i]['P'], wSorted, G[i][j]['A'], A_neighb, trunc, algo, G.graph['shelf'] )
      truncationInfoGlobal_update( truncInfoGlobal, infoSweep )
  
  else:					# graph with equivalent edges
    for i,j in G.edges():
      if G.node[i]['equivTo']==i and G.node[j]['equivTo'] in [i,j]:
        print ' --- i,j=',i,j,'---------'
        selfy = (G.node[j]['equivTo'] == i)
        k_minus_j = neighb_minus_j( G, i, j )
        wSorted   = w_sorted( G, i, j )
        A_neighb = []
        for k in k_minus_j:
          if G.node[k]['equivTo'] == i or (G.node[k]['equivTo'] == j and selfy ):
            A_neighb.append( G[i][j]['A'] )
          else:
            A_neighb.append( G[G.node[k]['equivTo']][i]['A'] )
        G[i][j]['B'], infoSweep = evolve_edge( G.node[i]['P'], wSorted, G[i][j]['A'], A_neighb, trunc, algo, G.graph['shelf'] )
        truncationInfoGlobal_update( truncInfoGlobal, infoSweep )
  
  # Move all 'B' to 'A'
  for i,j in G.edges():
    if 'A' in G[i][j]:
      G[i][j]['A'] = G[i][j]['B']
      del G[i][j]['B']
