# MPEM toolkit v 1.0 by Thomas Barthel

import numpy as np
import math

from truncation import *


#__________________________________________________________________________________________
#================================================================ generic graph functions =

#---------------------------------------------------------------[ totalNoNodes ]-
def totalNoNodes( G ):
  if 'equivTo' not in G.node[0]:
    return G.number_of_nodes()
  return sum([G.node[i]['NoEquiv'] for i in G.nodes() if G.node[i]['equivTo']==i])

#---------------------------------------------------------------[ NoEquivNodes ]-
def NoEquivNodes( G, i ):
  if 'equivTo' not in G.node[0]:
    return 1
  return G.node[G.node[i]['equivTo']]['NoEquiv']

#----------------------------------------------------------------[ dimension_i ]-
def dimension_i( G, i ):
  return len( G.node[i]['P'] )

#-------------------------------------------------------------[ neighb_minus_j ]-
def neighb_minus_j( G, i, j ):
  neighb = sorted(G.neighbors(i))
  neighb.remove(j)
  return neighb


#__________________________________________________________________________________________
#============================================================== Transition matrix w(s|s') =

#-------------------------------------------------------------------[ w_sorted ]-
def w_sorted( G, i, j ):
  w = G.node[i]['w']
  neighb = sorted(G.neighbors(i))
  idx_j = neighb.index(j)
  idx_E = range(len(neighb))
  idx_E.remove( idx_j )
  return w.transpose( [0,idx_j+1]+[idx_e+1 for idx_e in idx_E] )


#--------------------------------------------------------------------[ probe_w ]-
# check whether w is indeed a stochastic matrix
def probe_w( w ):
  assert w.ndim >= 2
  assert max(w.flatten()) >= 0
  W = w.reshape( w.shape[0], -1 )
  for sv in range(W.shape[1]):
    if abs(sum(W[:,sv])-1) > 1.e-8:
      print 'Problem in column sv =', s
      print 'of the transition matrix', W
      assert False


#-------------------------------------------------------------[ w_matixProduct ]-
# bring the transition matrix w(si|s1's2'..) into MP form
#  w(si|s1's2'..) = A(si,s1')*A(s2')*...*A(sz')
def w_matrixProduct( w ):
  assert w.ndim >= 3
  noTrunc = NoTruncator()
  A = (w.ndim-2)*[0,]
  d = w.shape
  X = w.reshape( np.prod(d), 1 )
  for k in range(3,w.ndim)[::-1]:
    X = X.reshape( X.shape[0]/d[k], d[k]*X.shape[1] )
    print A,k
    X,L,A[k-2], info = truncate( X, noTrunc )
    A[k-2] = A[k-2].reshape(len(L),d[k],-1)
    X = np.einsum('aq,q->aq',X,L)
  A[0] = X.reshape(d[0],d[1],d[2],-1)
  return A


#-----------------------------------------------------------------[ w_majority ]-
# Transition matrix for majority rule
def w_majority_z( z, d ):
  if d!=2:
    print ' d is not 2! Reset d accordingly!'
    assert False
  magn=[-1,1]
  w = np.zeros((z+1)*[d,])
  for Sj in np.ndindex(z*(d,)):
    Sum_Sj = sum([magn[sj] for sj in Sj])
    if Sum_Sj == 0:
      w[(0,)+Sj] = 0.5
      w[(1,)+Sj] = 0.5
    elif Sum_Sj>0:
      w[(1,)+Sj] = 1
    else:
      w[(0,)+Sj] = 1
  return w

def w_majority( G, i, d ):
  return w_majority_z( d, G.degree(i) )


#------------------------------------------------------------------[ w_glauber ]-
# Transition matrix for Ising Glauber dynamics
def w_glauber_z( z, d, beta, J ):
  if d!=2:
    print ' d is not 2! Reset d accordingly!'
    assert False
  magn=[-1,1]
  w = np.zeros((z+1)*[d,])
  for Sj in np.ndindex(z*(d,)):
    Sum_JSj = sum([J*magn[sj] for sj in Sj])
    Z = math.exp(-beta*Sum_JSj) + math.exp(beta*Sum_JSj)
    w[(0,)+Sj] = math.exp(-beta*magn[0]*Sum_JSj) / Z
    w[(1,)+Sj] = math.exp(-beta*magn[1]*Sum_JSj) / Z
  return w

def w_glauber( G, i, d, beta, J ):
  return w_glauber_z( G.degree(i), d, beta, J )

