# MPEM toolkit v 1.0 by Thomas Barthel

from math import sqrt
import numpy.linalg as linalg



#__________________________________________________________________________________________
#============================================================================ Truncations =

#------------------------------------------------------------------[ Truncator ]-
def Truncator(statesMin=1, statesMax=1e6, do_normalize=True, errorMax=1, schmidtTruncMax=1e-5):
  return {'statesMin':statesMin, 'statesMax':statesMax, 'do_normalize':do_normalize, 'errorMax':errorMax, 'schmidtTruncMax':schmidtTruncMax}

def NoTruncator():
  return {'statesMin':1, 'statesMax':1e6, 'do_normalize':True, 'errorMax':1, 'schmidtTruncMax':1e-14}

def truncatorLabel( trunc ):
  ret = ''
  if trunc['statesMin'] != 1:    ret += 'Mmin' + str(trunc['statesMin'])
  if trunc['statesMax'] != 1e6:  ret += 'Mmax' + str(trunc['statesMax'])
  if trunc['errorMax']  != 1:    ret += 'eps'  + str(trunc['errorMax'])
  if trunc['schmidtTruncMax'] != 1:  ret += 'sTrunc' + str(trunc['schmidtTruncMax'])
  return ret


#-------------------------------------------------------[ truncationInfo*_init ]-
def truncationInfoSweep_init():
  infoSweep= {}
  infoSweep['statesKept_max']   = 0
  infoSweep['schmidtTrunc_max'] = 0
  infoSweep['error_max']        = 0
  infoSweep['error_sum']        = 0
  infoSweep['cost_max']         = 0
  infoSweep['cost_sum']         = 0
  return infoSweep

def truncationInfoGlobal_init():
  infoGlobal = {}
  infoGlobal['statesKept_max']   = 0
  infoGlobal['schmidtTrunc_max'] = 0
  infoGlobal['error_max']        = 0
  infoGlobal['error_sum_max']    = 0
  infoGlobal['cost_max']        = 0
  infoGlobal['cost_sum']        = 0
  return infoGlobal


#-----------------------------------------------------[ truncationInfo*_update ]-
# truncation info has fields 'statesKept', 'schmidtTruncMax', 'error', 'cost'
def truncationInfoSweep_update( infoSweep, info ):
  if len(infoSweep) == 0:
    infoSweep.update( truncationInfoSweep_init() )
  infoSweep['statesKept_max']   = max( info['statesKept'],   infoSweep['statesKept_max'] )
  infoSweep['schmidtTrunc_max'] = max( info['schmidtTruncMax'], infoSweep['schmidtTrunc_max'] )
  infoSweep['error_max']        = max( info['error'],        infoSweep['error_max'] )
  infoSweep['error_sum']       += info['error']
  infoSweep['cost_max']         = max( info['cost'],         infoSweep['cost_max'] )
  infoSweep['cost_sum']        += info['cost']

def truncationInfoGlobal_update( infoGlobal, infoSweep ):
  if len(infoGlobal) == 0:
    infoGlobal.update( truncationInfoGlobal_init() )
  infoGlobal['statesKept_max']   = max( infoSweep['statesKept_max'],  infoGlobal['statesKept_max'] )
  infoGlobal['schmidtTrunc_max'] = max( infoSweep['schmidtTrunc_max'], infoGlobal['schmidtTrunc_max'] )
  infoGlobal['error_max']        = max( infoSweep['error_max'], infoGlobal['error_max'] )
  infoGlobal['error_sum_max']    = max( infoSweep['error_sum'], infoGlobal['error_sum_max'] )
  infoGlobal['cost_max']         = max( infoSweep['cost_max'],  infoGlobal['cost_max'] )
  infoGlobal['cost_sum']        += infoSweep['cost_sum']


#-------------------------------------------------------------------[ truncate ]-
# Do SVD C=U*L*V and truncate Schmidt coefficients L.
def truncate( C, trunc ):
  U,L,V = linalg.svd( C, full_matrices=False )
  norm = linalg.norm(L)
  L *= 1./norm
  
  kept = len(L)
  err = 0
  smallest = 0
  while kept > trunc['statesMin']:
    newErr = err + L[kept-1]**2
    newSmallest = L[kept-1]
    if kept <= trunc['statesMax'] and (sqrt(newErr) > trunc['errorMax'] or newSmallest > trunc['schmidtTruncMax']):
      break
    err, smallest = [newErr, newSmallest]
    kept -= 1
  
  U = U[:,:kept]
  L = L[:kept]
  V = V[:kept,:]
  L *= 1./linalg.norm(L)
  if not trunc['do_normalize']:
    L *= norm
  
  info = { 'statesKept':kept, 'schmidtTruncMax':smallest, 'error':sqrt(err), 'cost':(min(C.shape)**2)*max(C.shape) }
  return U, L, V, info


#----------------------------------------------------------------[ truncate_dm ]-
# Diagonalize rho=U*R*herm(U) and truncate density matrix eigenvalues R.
def truncate_dm( rho, trunc ):
  R,U = linalg.eigh( rho )
  idx = sorted( range(len(R)), key=lambda k: -R[k] )
  U = U[:,idx]
  R = R[idx]
  norm = sum(R)
  R *= 1./norm
  
  kept = len(R)
  err = 0
  smallest = 0
  while kept > trunc['statesMin']:
    if R[kept-1] > 0:
      newErr = err + R[kept-1]
      newSmallest = sqrt(R[kept-1])
      if kept <= trunc['statesMax'] and (sqrt(newErr) > trunc['errorMax'] or newSmallest > trunc['schmidtTruncMax']):
        break
      err, smallest = [newErr, newSmallest]
    kept -= 1
  
  U = U[:,:kept]
  R = R[:kept]
  R *= 1./sum(R)
  if not trunc['do_normalize']:
    R *= norm
  
  info = { 'statesKept':kept, 'schmidtTruncMax':smallest, 'error':sqrt(err), 'cost':rho.shape[0]**3 }
  return U, R, info
