#!/usr/bin/python
# MPEM toolkit v 1.0 by Thomas Barthel

import sys
from math import cos, sin, pi
import numpy as np
import networkx as nx
import pickle

from generica import *
from transitionMatrix import *


#============================================== Certain graphs ==

#--------------------------------------
def BetheLattice_equiv( z, h, equivInfo=True ):
  assert z>1 and h>0
  G = nx.Graph()
  n = 0
  alpha0 = 0
  for nh in range(h+1):
    NoEquiv = 1 if (nh == 0) else 3*(z-1)**(nh-1)
    n0 = n
    for ne in range(NoEquiv):
      G.add_node(n)
      if n0 > 0:  G.add_edge(n,n0+(ne-NoEquiv)/(z-1+(nh==1)))
      if equivInfo:
        G.node[n]['equivTo'] = n0
      alpha = ne*2*pi/NoEquiv + alpha0
      G.node[n]['pos'] = (nh*cos(alpha),nh*sin(alpha))
      n += 1
    if equivInfo:
      G.node[n0]['NoEquiv'] = NoEquiv
    if nh > 0:  alpha0 -= pi/NoEquiv/2
  return G

#--------------------------------------
# Here we exploit that many vertices (and edges) are equivalent and reduce the graph correspndingly.
def BetheLattice_equivTrunc( z, h ):
  assert z>1 and h>0
  G = nx.Graph()
  i = 0
  G.add_node(i);  G.node[i]['equivTo']=i;  G.node[i]['NoEquiv']=1;  G.node[i]['pos']=(0,0);  i+=1
  G.add_node(i);  G.node[i]['equivTo']=2;  G.node[i]['pos']=(0,1);  G.add_edge(i,0);  i+=1
  dAlpha = 0.5*pi/(z-1)
  i0 = 0
  for nh in range(1,h+1):
    j0 = i
    for ne in range(z-1):
      G.add_node(i)
      print 'edge',(i,i0)
      G.add_edge(i,i0)
      G.node[i]['equivTo'] = j0
      G.node[i]['pos'] = (sin(ne*dAlpha),-nh+1-cos(ne*dAlpha))
      i += 1
    G.node[j0]['NoEquiv'] = 3*(z-1)**(nh-1)
    i0 = i - (z-1)
  return G

#--------------------------------------
# Here we exploit that all vertices (and edges) are equivalent and reduce the graph correspndingly.
def BetheLattice_equivInfinite( z ):
  assert z>1
  G = nx.Graph()
  i = 0
  G.add_node(i);  G.node[i]['equivTo']=i;  G.node[i]['NoEquiv']=1;  G.node[i]['pos']=(0,0);  
  i+=1
  dAlpha = 2*pi/z
  for ne in range(z):
    G.add_node(i)
    G.add_edge(i,0)
    G.node[i]['equivTo'] = int(ne>0)
    G.node[i]['pos'] = (sin(ne*dAlpha),cos(ne*dAlpha))
    i += 1
  return G


#============================================= initializations ==

#--------------------------------------
def initialize_graph( graph, param ):
  if graph == 'randomRegular':
    G = nx.random_regular_graph( int(param['z']), long(param['Nn']), long(param['seed']) )
    label = graph + '_z%iNn%i' % ( param['z'], param['Nn'] )
  elif graph == 'balancedTree':
    G = nx.balanced_tree( int(param['r']), int(param['h']) )
    label = graph + '_r%ih%i' % ( param['r'], param['h'] )
  elif graph == 'Bethe':
    G = BetheLattice_equiv( int(param['z']), int(param['h']), equivInfo=True )
    label = graph + '_z%ih%i' % ( param['z'], param['h'] )
  elif graph == 'BetheTrunc':
    if param['h'] > 0:
      G = BetheLattice_equivTrunc( int(param['z']), int(param['h']) )
    else:
      G = BetheLattice_equivInfinite( int(param['z']) )
    label = graph + '_z%ih%i' % ( param['z'], param['h'] )
  elif graph == 'BetheInf':
    G = BetheLattice_equivInfinite( int(param['z']) )
    label = graph + '_z%i' % ( param['z'] )
  else:
    print '\nUnknown graph type',graph
    assert False
  
  isolated_nodes = nx.isolates(G)
  for i in isolated_nodes: G.remove_node(i)
  
  print 'Number of vertices =', G.number_of_nodes()
  if 'equivTo' in G.node[0]:
    print 'Total number of vertices =', sum([G.node[i]['NoEquiv'] for i in G.nodes() if G.node[i]['equivTo']==i])
  print 'maximum vertex degree =', max([G.degree(i) for i in G.nodes()])
  print 'vertex degrees =', [G.degree(i) for i in G.nodes()]
  for i in G.nodes():
    print i,G.node[i]
  
  return G, label


#--------------------------------------
def initialize_model( G, trans, tparam, initial, iparam ):
  d = 2  # number of local states
  if trans == 'Glauber':
    tlabel = trans + '_beta%gJ%g' % ( tparam['beta'], tparam['J'] )
  elif trans == 'majority':
    tlabel = trans + 'majority'
  else:
    print '\nUnknown transition type',trans
    assert False
  
  if initial == 'homogene':
    P0 = np.array( [0.5-iparam['bias']/2.,0.5+iparam['bias']/2.] )
    ilabel = initial + '_bias%g' % iparam['bias']
  else:
    print '\nUnknown initial distr. type',initial
    assert False
  
  for i in G.nodes():
    G.node[i]['P'] = P0
    if trans == 'Glauber':
      G.node[i]['w'] = w_glauber(G,i,d,tparam['beta'],tparam['J'])
    elif trans == 'majority':
      G.node[i]['w'] = w_majority(G,i,d)
    probe_w( G.node[i]['w'] )
  
  return d, tlabel, ilabel



#=================================================================================
#========================================================================={ MAIN ]
graphType  = 'BetheTrunc'
graphParam = {}
transType  = 'Glauber'
transParam = {}
initialType  = 'homogene'
initialParam = {}
out_model  = ''

program = sys.argv[0].rsplit('.py',1)[0].rsplit('/',1)[-1]
print '\n >>>> Creates graph, initial state, and transition matrices <<<<  v 1.0, T. Barthel'
print 'usage: ' + program + '.py'
print '\t(-graphType <.='+graphType+'>)  (-graphParam <label> <value> ..)'
print '\t(-transitionType <.='+transType+'>)  (-transitionParam <label> <value> ..)'
print '\t(-initialType <.='+initialType+'>)  (-initialParam <label> <value> ..)'
print '\t(-out_model <*.model.bin =std string>)'
print ' Available graph types (and required parameters):  randomRegular (z, Nn, seed), balancedTree (r, h), Bethe (z, h), BetheTrunc (z, h), BetheInf (z).'
print '  BetheTrunc with parameters (z,h=-1)  ==  BetheInf with (z).'
print ' Available transition types (and required parameters):  Glauber (beta, J), majority .'
print ' Available initial distr. types (and required parameters):  homogene (bias).'
print ' The result is stored in pickle format.\n'

if len(sys.argv) == 1:
  sys.exit(0)


# ----------------------------------------- read command line parameters
modable = set(['-graphParam','-transitionParam','-initialParam'])
mode = 'none'
n = 1

while n < len(sys.argv) and ( is_cmdlineArg( sys.argv[n] ) or mode!='none' ):
  arg = sys.argv[n]
  [dn, mode, switch] = read_cmdlineArg( arg, mode, modable )
  n += dn
  
  if switch == '-graphType':
    graphType = sys.argv[n]
  elif switch == '-graphParam':
    graphParam[ sys.argv[n] ] = float(sys.argv[n+1])
    n += 1
  elif switch == '-transitionType':
    transType = sys.argv[n]
  elif switch == '-transitionParam':
    transParam[ sys.argv[n] ] = float(sys.argv[n+1])
    n += 1
  elif switch == '-initialType':
    initialType = sys.argv[n]
  elif switch == '-initialParam':
    initialParam[ sys.argv[n] ] = float(sys.argv[n+1])
    n += 1
  elif switch == '-out_model':
    out_model = sys.argv[n]
  else:
    print '\nUnknown switch ',switch
    sys.exit(2)
  n += 1


# ----------------------------------------- GENERATE GRAPH
G, graph = initialize_graph( graphType, graphParam )


# ----------------------------------------- TRANSITION MATRICES and INITIAL DISTRIB.
d, trans, initial = initialize_model( G, transType, transParam, initialType, initialParam )

G = nx.DiGraph(G)


# ----------------------------------------- OUTPUT FILE
if len(out_model) == 0:
  out_model = './'+graph+'~'+trans+'~'+initial+'.model.bin'
print '\noutfile =', out_model

G.graph['d'] = d
nx.write_gpickle( G, out_model )

print ''
